%skeleton "lalr1.cc"
%require "3.0.4"

%define parser_class_name {Parser}
%define api.namespace {parser}

%code requires
{
	#include<vector>
	#include<sstream>
	#include "../asp-domain/asp-domain.hpp"
	using namespace domain;
	namespace parser {
		class Driver;
		class FlexScanner;
	}
}

%parse-param { parser::Driver& driver }
%parse-param { parser::FlexScanner& scanner }
%lex-param   { parser::FlexScanner& scanner }
%locations
%error-verbose

%code
{
	#include "Driver.h"
	// Prototype for the yylex function
	static int yylex(parser::Parser::semantic_type * yylval, parser::Parser::location_type* yylloc, parser::FlexScanner &scanner);
}

%union
{
	std::string* string;
	Term* term;
	vector<Term *>* terms;
	ClassicalLiteral* classicalLiteral;
	vector<TermTuple *>* termTuples;
	TermTuple* termTuple;
	Program* program;
	Statement* statement;
	OptimizeStatement* optimize;
	BodyElement* bodyElement;
	OptimizeElement* optimizeElement;
	WeightAtLevel* weightAtLevel;
	RuleHead* head;
	NafLiteral* nafLiteral;
	ChoiceElement* choiceElement;
	AggregateElement* aggregateElement;
	pair<std::string*, TermTuples*>* nameTuplesP;
	vector<BodyElement *>* bodyElements;
	vector<ClassicalLiteral *>* classicalLiterals;
	vector<ChoiceElement *>* choiceElements;
	vector<NafLiteral*>* nafLiterals;
	vector<OptimizeElement *>* optimizeElements;
	vector<AggregateElement *>* aggregateElements;
	Aggregate* aggregate;
};

%token          END        0 "end of file"

%token		NAF "not"
%token		CONS ":-"
%token		WCONS ":~"
%token		UNEQUAL
%token		LESS_OR_EQ "<="
%token		GREATER_OR_EQ ">="
%token		INTERVAL ".."
%token		EXP "**"
%token		AGGREGATE_COUNT "#count"
%token		AGGREGATE_MAX "#max"
%token		AGGREGATE_MIN "#min"
%token		AGGREGATE_SUM "#sum"
%token		MINIMIZE
%token		MAXIMIZE

%token	<string>	ID
%token	<string>	VARIABLE
%token	<string>	STRING
%token	<string>	NUMBER
%token	<string>	METASTATEMENT

%left	".."	// actually, association does not matter.
%left	'^'
%left	'?'
%left	'&'
%left	'+' '-'
%left	'*' '/' '\\'	// multiplication and division bind stronger than plus/minus. all four are left-associative (meaning a+b+c will be interpreted as (a+b)+c ).
%right	"**"
%left	UNARYMINUS '~' // UNARYMINUS: fictious symbol for precedence of unary minus symbol. cmp. Bison user guide chapter 5.4.

%type	<string>	binop optimize_function aggregate_function
%type	<term>	 	term
%type	<terms>	pooling_term termtuple
%type	<nafLiteral>	builtin_atom naf_literal
%type	<termTuples>	termtuples
%type	<nameTuplesP>	positive_classical_literal // since positive_classical_literal is also used for functions and symbolic constants
%type	<classicalLiteral>	classical_literal /*query*/
%type	<statement>	statement
%type	<program>	program
%type	<bodyElement>	body_element
%type	<bodyElements>	body
%type	<head>		head choice
%type	<weightAtLevel>	weight_at_level
%type	<optimize>	optimize
%type	<classicalLiterals>	disjunction
%type	<choiceElements>	choice_elements unbounded_choice
%type	<choiceElement>	choice_element
%type	<nafLiterals>	naf_literals
%type	<optimizeElements>	optimize_elements
%type	<optimizeElement>	optimize_element
%type	<aggregateElement>	aggregate_element
%type	<aggregateElements>	aggregate_elements
%type	<aggregate>	unbounded_aggregate aggregate
%%

%start start;

start:	program	{ driver.program = $1; }
     ;

program: /* empty */		{ $$ = new Program(); }
       | program statement	{ $$ = $1; $$->statements->push_back($2); }
       | program METASTATEMENT	{ $$ = $1; $$->metastatements->push_back($2); }
//       | program query //FIXME: only one query per program allowed
       ;

/* TODO ambiguity with arithmetic operator ?
 * (consider the input h ? 1 < {}. )
query: classical_literal '?'	{ $$ = $1; }
     ;
*/
     
statement: ":-" '.'					{ $$ = new Rule(new Disjunction(new vector<ClassicalLiteral*>()), new Body(new vector<BodyElement*>())); }
         | ":-" body '.'				{ $$ = new Rule(new Disjunction(new vector<ClassicalLiteral*>()), new Body($2)); }
         | head '.'					{ $$ = new Rule($1, new Body(new vector<BodyElement*>())); }
         | head ":-" '.'				{ $$ = new Rule($1, new Body(new vector<BodyElement*>())); } // semantically the same as above (though gringo does not accept it)
         | head ":-" body '.'				{ $$ = new Rule($1, new Body($3)); }
         | ":~" '.' '[' weight_at_level ']'		{ $$ = new WeakConstraint(new Body(new vector<BodyElement*>()), $4); }
         | ":~" body '.' '[' weight_at_level ']'	{ $$ = new WeakConstraint(new Body($2), $5); }
         | optimize '.'					{ $$ = $1; }
         ;
         
head: disjunction	{ $$ = new Disjunction($1); }
    | choice		{ $$=$1; }
    ;
    
body: body_element		{ $$ = new vector<BodyElement*>(); $$->push_back($1); }
    | body ',' body_element	{ $$=$1; $$->push_back($3); }
    ;

body_element: naf_literal	{ $$ = $1; }
            | aggregate		{ $$ = $1; }
            | "not" aggregate	{ $2->nafNegated=true; $$ = $2; } // semantic: strong negation
            ;
            
disjunction: classical_literal			{ $$ = new vector<ClassicalLiteral *>(); $$->push_back($1); }
           | disjunction '|' classical_literal	{ $$=$1; $$->push_back($3); }
           ;
           
                
choice: unbounded_choice	{ $$ = new Choice($1); }
      | term binop unbounded_choice	{ $$ = new Choice($3, new Guard($2, $1, true)); } // experimental result: gringo admits pooling of guards only within parentheses
      | unbounded_choice binop term	{ $$ = new Choice($1, new Guard($2, $3, false)); }
      | term binop unbounded_choice binop term	{ $$ = new Choice($3, new Guard($2, $1, true), new Guard($4, $5, false)); }
      ;
      
unbounded_choice: '{' '}'			{ $$ = new vector<ChoiceElement *>(); }
                | '{' choice_elements '}'	{ $$=$2; }
                ;
choice_elements: choice_element				{ $$ = new vector<ChoiceElement *>(); $$->push_back($1); }
               | choice_elements ';' choice_element	{ $$=$1; $$->push_back($3); }
               ;
               
choice_element: classical_literal			{ $$ = new ChoiceElement($1, new LiteralBody()); }
              | classical_literal ':'			{ $$ = new ChoiceElement($1, new LiteralBody()); }
              | classical_literal ':' naf_literals	{ $$ = new ChoiceElement($1, new LiteralBody($3)); }
              ;
              
aggregate: unbounded_aggregate				{ $$ = $1; }
         | term binop unbounded_aggregate		{ $$ = $3; $$->guard_l = new Guard($2, $1, true); } // experimental result: gringo admits pooling of guards only within parentheses
         | unbounded_aggregate binop term		{ $$ = $1; $$->guard_r = new Guard($2, $3, false); }
         | term binop unbounded_aggregate binop term	{ $$ = $3; $$->guard_l = new Guard($2, $1, true); $$->guard_r = new Guard($4, $5, false); }
         ;
              
unbounded_aggregate: aggregate_function '{' aggregate_elements '}'	{ $$ = new Aggregate($1, false, $3); } // TODO: actually, #aggr{} would also be a direct unbounded_aggregate according to the ASP grammar, but that would introduce a grammar ambiguity and a shift/reduce conflict, since aggregate_elements may be the empty string as well.
                   ;
                   
aggregate_elements: aggregate_element				{ $$ = new vector<AggregateElement *>(); $$->push_back($1); }
                  | aggregate_elements ';' aggregate_element	{ $$=$1; $$->push_back($3); }
                  ;
                  
aggregate_element: /* empty */			{ $$ = new AggregateElement(new TermTuple(new vector<Term*>())); }	// TODO: error in ASP syntax? Gringo does not accept.
                 | termtuple			{ $$ = new AggregateElement(new TermTuple($1)); }
                 | termtuple ':'		{ $$ = new AggregateElement(new TermTuple($1)); }
                 | termtuple ':' naf_literals	{ $$ = new AggregateElement(new TermTuple($1), new LiteralBody($3)); }
                 | ':'				{ $$ = new AggregateElement(new TermTuple(new vector<Term*>())); }	// TODO: error in ASP syntax? Gringo does not accept.
                 | ':' naf_literals		{ $$ = new AggregateElement(new LiteralBody($2)); }
                 ;
                 
aggregate_function: "#count"	{ $$ = new std::string("#count"); }
                  | "#max"	{ $$ = new std::string("#max"); }
                  | "#min"	{ $$ = new std::string("#min"); }
                  | "#sum"	{ $$ = new std::string("#sum"); }
                  ;
                  
optimize: optimize_function '{' '}'			{ $$ = new OptimizeStatement($1, new vector<OptimizeElement *>()); }
        | optimize_function '{' optimize_elements '}'	{ $$ = new OptimizeStatement($1, $3); }
        ;
        
optimize_elements: optimize_element				{ $$ = new vector<OptimizeElement *>(); $$->push_back($1); }
                 | optimize_elements ';' optimize_element	{ $$=$1; $$->push_back($3); }
                 ;
                 
optimize_element: weight_at_level			{ $$ = new OptimizeElement($1, new LiteralBody()); }
                | weight_at_level ':'			{ $$ = new OptimizeElement($1, new LiteralBody()); }	// ':' syntactically and semantically unneccessary
                | weight_at_level ':' naf_literals	{ $$ = new OptimizeElement($1, new LiteralBody($3)); }
                ;
                
optimize_function: MINIMIZE	{ $$ = new std::string("#minimize"); }
                 | MAXIMIZE	{ $$ = new std::string("#maximize"); }
                 ;
                 
weight_at_level: term				{ $$ = new WeightAtLevel($1); }	// weight only (level = 0)
               | term '@' term			{ $$ = new WeightAtLevel($1, $3); }	// weight @ level
               | term ',' termtuple		{ $$ = new WeightAtLevel($1, new TermTuple($3)); }	// weight (level = 0), termtuple
               | term '@' term ',' termtuple	{ $$ = new WeightAtLevel($1, $3, new TermTuple($5)); }	// weight @ level, termtuple
               ;
               
naf_literals: naf_literal			{ $$ = new vector<NafLiteral*>(); $$->push_back($1); }
            | naf_literals ',' naf_literal	{ $$=$1; $$->push_back($3); }
            ;
            
naf_literal: classical_literal	{ $$ = $1; }
           | builtin_atom	{ $$ = $1; }
           | "not" classical_literal	{ $2->nafNegated=true; $$ = $2;}
           ;

classical_literal: positive_classical_literal		{ $$ = new ClassicalLiteral($1->first, false, false, $1->second); delete $1; }
                 | '-' positive_classical_literal	{ $$ = new ClassicalLiteral($2->first, true, false, $2->second); delete $2; }
                 ;
                 
positive_classical_literal: ID				{ $$ = new pair<std::string*, TermTuples*>($1, NULL); }
                          | ID '(' ')'			{ $$ = new pair<std::string*, TermTuples*>($1, NULL); } // experimental result with gringo: semantically the same as without parentheses.
                          | ID '(' termtuples ')'	{ { $$ = new pair<std::string*, TermTuples*>($1, new TermTuples($3)); } } // experimental result with gringo: ',' (tuple generation) binds stronger than ';' (pooling)
                          ;
                          
termtuples: termtuple	{ $$ = new vector<TermTuple *>(); $$->push_back(new TermTuple($1)); }
          | termtuples ';' termtuple {$1->push_back(new TermTuple($3)); $$=$1; }
          ;

builtin_atom: term binop term	{ $$ = new BuiltinAtom($1, $2, $3); }
/*             | pooling_term binop term */ // experiment with gringo: "test :- 5;8 < 6." -> syntax error. -> only nonpooled term before binop. also after binop.
            ;
            
binop: '='	{ $$ = new std::string("="); }
     | UNEQUAL	{ $$ = new std::string("!="); }
     | '<'	{ $$ = new std::string("<"); }
     | '>'	{ $$ = new std::string(">"); }
     | "<="	{ $$ = new std::string("<="); }
     | ">="	{ $$ = new std::string(">="); }
     ;
     
termtuple: term			{ $$ = new vector<Term *>(); $$->push_back($1); }
	 | termtuple ',' term	{ $1->push_back($3); $$=$1; }
	 ;
     
// term may include pools, but only in parentheses. E.g. "1;2;3" is a pooling_term and a term, but not a term, whereas "(1;2;3)" is.
term: positive_classical_literal	{ if($1->second == NULL) {$$ = new Constant($1->first);} else {$$ = new Function($1->first, $1->second);} delete $1; } // actually function
               | NUMBER		{ $$ = new Constant($1); }
               | STRING		{ $$ = new Constant($1); }
               | VARIABLE	{ $$ = new Variable($1); }
               | '_'		{ $$ = new AnonymousVariable(); }
               | '(' term ')'	{ $$ = new BracketedTerm($2); }
               | '(' pooling_term ')'	{ $$ = new PoolingTerm($2); }
               | '-' term %prec UNARYMINUS	{ $$ = new UnaryTerm($2, NEGATION); }
               | '~' term	{ $$ = new UnaryTerm($2, COMPLEMENT); }
               | '|' term '|'	{ $$ = new UnaryTerm($2, ABSOLUTE); }
               | term '+' term	{ $$ = new BinaryTerm($1, new std::string("+"), $3); }
               | term '-' term	{ $$ = new BinaryTerm($1, new std::string("-"), $3); }
               | term '*' term	{ $$ = new BinaryTerm($1, new std::string("*"), $3); }
               | term '/' term	{ $$ = new BinaryTerm($1, new std::string("/"), $3); }
               | term '\\' term	{ $$ = new BinaryTerm($1, new std::string("\\"), $3); }
               | term '&' term	{ $$ = new BinaryTerm($1, new std::string("&"), $3); }
               | term '?' term	{ $$ = new BinaryTerm($1, new std::string("?"), $3); }
               | term '^' term	{ $$ = new BinaryTerm($1, new std::string("^"), $3); }
               | term "**" term	{ $$ = new BinaryTerm($1, new std::string("**"), $3); }
               | term ".." term	{ $$ = new BinaryTerm($1, new std::string(".."), $3); } // integer interval
               ;
               
pooling_term: term ';' term	{ $$ = new vector<Term *>(); $$->push_back($1); $$->push_back($3); }
            | pooling_term ';' term	{ $1->push_back($3); $$=$1; }
            ;

%%

void parser::Parser::error(const parser::Parser::location_type& l, const std::string& m)
{
	driver.error(l, m);
}

// Now that we have the Parser declared, we can declare the Scanner and implement
// the yylex function
#include "scanner.h"
static int yylex(parser::Parser::semantic_type * yylval, parser::Parser::location_type* yylloc, parser::FlexScanner &scanner) {
	return scanner.yylex(yylval, yylloc);
}
